
/* Name: main.c
 * Project: AVR USB driver for CDC interface on Low-Speed USB
 *              for ATtiny2313
 * Author: Osamu Tamura
 * Creation Date: 2007-10-03
 * Tabsize: 4
 * Copyright: (c) 2007-2009 by Recursion Co., Ltd.
 * License: Proprietary, free under certain conditions. See Documentation.
 *
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "oddebug.h"


/*
enum {
    SEND_ENCAPSULATED_COMMAND = 0,
    GET_ENCAPSULATED_RESPONSE,
    SET_COMM_FEATURE,
    GET_COMM_FEATURE,
    CLEAR_COMM_FEATURE,
    SET_LINE_CODING = 0x20,
    GET_LINE_CODING,
    SET_CONTROL_LINE_STATE,
    SEND_BREAK
};
*/

//#define __DEBUG__BITNAME__

#define HIDSERIAL_INBUFFER_SIZE 32


static uchar received = 0;
static uchar inBuffer[HIDSERIAL_INBUFFER_SIZE+1];
static uchar recv_len=0;


PROGMEM const char usbHidReportDescriptor[] = {
  0x06, 0xa0, 0xff, // USAGE_PAGE (Vendor Defined Page 1)
  0x09, 0x01,       // USAGE (Vendor Usage 1)
  0xa1, 0x01,       // COLLECTION (Application)

  // Input Report
  0x09, 0x02,       // Usage ID - vendor defined
  0x15, 0x00,       // Logical Minimum (0)
  0x26, 0xFF, 0x00, // Logical Maximum (255)
  0x75, 0x08,       // Report Size (8 bits)
  0x95, 0x08,       // Report Count (8 fields), must be 8
  0x81, 0x02,       // Input (Data, Variable, Absolute)

  // Output report
  0x09, 0x03,       // Usage ID - vendor defined
  0x15, 0x00,       // Logical Minimum (0)
  0x26, 0xFF, 0x00, // Logical Maximum (255)
  0x75, 0x08,       // Report Size (8 bits)
  0x95, HIDSERIAL_INBUFFER_SIZE,       // Report Count (16 fields)
  0x91, 0x02,       // Output (Data, Variable, Absolute)

  0xc0              // END_COLLECTION
};


#define	TX_SIZE		HIDSERIAL_INBUFFER_SIZE
#define	TX_MASK		(TX_SIZE-1)
#define	RX_SIZE		HIDSERIAL_INBUFFER_SIZE

/* command buffer */
static uchar    rx_buf[RX_SIZE+1], tx_buf[TX_SIZE+1];
static uchar    iwptr, uwptr;
static uchar    tos, nos;


/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */


usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	return 0;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionRead                                                           */
/*---------------------------------------------------------------------------*/

uchar usbFunctionRead(uchar *data, uchar len)
{
	return 0;
}

void usbFunctionWriteOut(uchar *data, uchar len)
{
uchar i,c;
	if(len>0) {
		if( (*data==0x0d) || (*data==0x0a) )
			return;
		if(len>HIDSERIAL_INBUFFER_SIZE) len=HIDSERIAL_INBUFFER_SIZE;
		i=0;
		while(i<len) {
			c=data[i];
			inBuffer[i]=c;
			i++;
			if(c==0) break;
		}
		received=1;
		recv_len=i;
	}
}


#ifdef __AVR_ATtiny4313__
static uchar	nos2;
// output bit data with clock function
static uchar 	outdata=0;
static uchar	*outport=NULL;
static uchar	outbport=8;
#ifdef __DEBUG__BITNAME__
char lbuff[17],bbuff[8];
#endif
#endif


static uchar u2h( uchar u )
{
	if( u>9 )
		u	+= 7;
	return u+'0';
}

static uchar h2u( uchar h )
{
	h	-= '0';
	if( h>9 )
		h	-= 7;
    return h;
}

#ifdef __AVR_ATtiny4313__
void return_error(void)
{
	rx_buf[0]	= 'E';
	iwptr	= 1;
}
#endif

void Translate_buffer(uchar *data)
{

    /*  postpone receiving next data    */
    usbDisableAllRequests();

#ifdef __AVR_ATtiny4313__
	uchar namelen=0;
static	uchar	bport=8;
#endif

    /*    host => device : Request  */
    do {
        char    c;

        //    delimiter?
        c    = *data++;
        if( c>0x20 ) {
            if( c & 0x40 )
                c   &= 0x5f;        //    to upper case
#ifdef __AVR_ATtiny4313__
			if( c == '.' ) namelen=uwptr;
#endif
            tx_buf[uwptr++]    = c;
            uwptr   &= TX_MASK;
            continue;
        }

        //    command
        if( uwptr==1 ) {
            volatile uchar  *addr = (uchar *)((unsigned int)tos);
            uchar   x;

           	//    get
			if( tx_buf[0]=='?' ) {
#ifdef __AVR_ATtiny4313__
				if( bport==8 ) {
#endif				
                x   = *addr;
                rx_buf[0]   = u2h(x>>4);
                rx_buf[1]   = u2h(x&0x0f);
				iwptr	= 2;
#ifdef __AVR_ATtiny4313__
				}
				else {
					rx_buf[0] = *addr & (1<<bport) ? '1' : '0';
					iwptr	= 1;
				}
#endif
			}
            //    set
			else if( tx_buf[0]=='=' ) {
#ifdef __AVR_ATtiny4313__
				if( bport==8 ) {
#endif				
                cli();
                *addr   = nos;
                sei();
#ifdef __AVR_ATtiny4313__
				}
				else {
					if( nos )
						*addr |= (1<<bport);
						else
						*addr &= ~(1<<bport);
				}
#endif
			}
            //    xor & set
			else if( tx_buf[0]=='^' ) {
#ifdef __AVR_ATtiny4313__
				if( bport==8 ) {
#endif
                cli();
                *addr   ^= nos;
                sei();
#ifdef __AVR_ATtiny4313__
				} else return_error();
#endif
			}
			//	 and & set
			else if( tx_buf[0]=='&' ) {
#ifdef __AVR_ATtiny4313__
				if( bport==8 ) {
#endif
				cli();
				*addr	&= nos;
				sei();
#ifdef __AVR_ATtiny4313__
				} else return_error();
#endif
			}
			//	or & set
			else if( tx_buf[0]=='|' ) {
#ifdef __AVR_ATtiny4313__
				if( bport==8 ) {
#endif
				cli();
				*addr	|= nos;
				sei();
#ifdef __AVR_ATtiny4313__
				} else return_error();
#endif
			}
#ifdef __AVR_ATtiny4313__
			// info
			else if( tx_buf[0]=='@' ) {
			    strncpy_P((char*)rx_buf,PSTR("cdc-io"),6);
				iwptr	= 6;
			}
			// set word
			else if( tx_buf[0]=='$' ) {
				if( bport==8 ) {
			 	cli();
				*addr	= nos2;
				*(addr+1)= nos;
				sei();
				} else return_error();
			}
			// reset interrupt flag
			else if( tx_buf[0]=='>' ) {
				if( bport!=8 ) {
				cli();
				outport		= (uchar *)((unsigned int)tos);
				outbport	= bport;
				outdata		= nos;
				sei();
				} else return_error();
			}
#ifdef __DEBUG__BITNAME__
			// debug
			else if( tx_buf[0]=='L') {
				strncpy((char*)rx_buf,lbuff,7);
				iwptr=7;
			}
			else if( tx_buf[0]=='O') {
				strncpy((char*)rx_buf,bbuff,3);
				iwptr=3;
			}
#endif
			else return_error();

			if( iwptr==0 ) {
				rx_buf[0]='P';
				iwptr=1;
			}
#endif
        }

        //    number
        else {
#ifdef __AVR_ATtiny4313__
		nos2 = nos;
#endif
		nos = tos;
#ifdef __AVR_ATtiny4313__
		bport=8;
		// support reg.0
		if(namelen==0) namelen=uwptr;
#ifdef __DEBUG__BITNAME__
		for(uchar k=0;k<namelen;k++) lbuff[k]=tx_buf[k];
#endif
		// byte value or port
		if( namelen==2 ) {
#endif
            tos = (h2u(tx_buf[0])<<4) | h2u(tx_buf[1]);
#ifdef __AVR_ATtiny4313__
        }
		// register name 4 byte
		else if( namelen==4 ) {
			tos=0x15; //GPIOR2;
			if( tx_buf[0]=='P' ) {
				// PIN
				if( (tx_buf[1]=='I')&&(tx_buf[2]=='N') ) {
					if( tx_buf[3]=='A' ) tos=0x19; //PINA;
					else if( tx_buf[3]=='B' ) tos=0x16; //PINB;
					else if( tx_buf[3]=='D' ) tos=0x10; //PIND;
				}

			}
			else if( tx_buf[0]=='D' ) {
				// DDR
				if( (tx_buf[1]=='D')&&(tx_buf[2]=='R') ) {
					if( tx_buf[3]=='A' ) tos=0x1a; // DDRA;
					else if( tx_buf[3]=='B' ) tos=0x17; // DDRB;
					else if( tx_buf[3]=='D' ) tos=0x11; //DDRD;
				}
				// DIDR
				else if( strncmp_P((char*)&tx_buf[1],PSTR("IDR"),3)==0 )
					tos=0x01; //DIDR;
			}
			else if( tx_buf[0]=='S' ) {
				// SREG
				if( strncmp_P((char*)&tx_buf[1],PSTR("REG"),3)==0 )
					tos=0x3f; // SREG;
			}
			else if( tx_buf[0]=='A' ) {
				// ACSR
				if( strncmp_P((char*)&tx_buf[1],PSTR("CSR"),3)==0 )
					tos=0x08; //ACSR;
			}
			else if( tx_buf[0]=='E' ) {
				if( tx_buf[1]=='E') {
				// EEAR
				if( strncmp_P((char*)&tx_buf[2],PSTR("AR"),2)==0 )
					tos=0x1e; //EEAR;
				// EEDR
				else if( strncmp_P((char*)&tx_buf[2],PSTR("DR"),2)==0 )
					tos=0x1d; //EEDR;
				// EECR
				else if( strncmp_P((char*)&tx_buf[2],PSTR("CR"),2)==0 )
					tos=0x1c; //EECR;
				}
			}
			else if( tx_buf[0]=='G' ) {
				// GIFR
				if( strncmp_P((char*)&tx_buf[1],PSTR("IFR"),3)==0 )
					tos=0x3a; //GIFR;
			}
			tos+=0x20;
		}
		// register name 5 byte
		else if( namelen==5) {
			tos=0x15; //GPIOR2;
			if( tx_buf[0]=='P' ) {
				// PORTx
				if( tx_buf[1]=='O' ) {
					if( strncmp_P((char*)&tx_buf[2],PSTR("RT"),2)==0 ) {
						if( tx_buf[4]=='A' ) tos=0x1b; //PORTA;
						else if( tx_buf[4]=='B' ) tos=0x18; //PORTB;
						else if( tx_buf[4]=='D' ) tos=0x12; //PORTD;
					}
				// PCMSK
				} else if( tx_buf[1]=='C' ) {
					if( strncmp_P((char*)&tx_buf[2],PSTR("MSK"),3)==0 )
						tos=0x20; //PCMSK;
				}
			}
			else if( tx_buf[0]=='M' ) {
				// MCUCR
				if( strncmp_P((char*)&tx_buf[1],PSTR("CUCR"),4)==0 )
					tos=0x35; //MCUCR;
			}
			else if( tx_buf[0]=='O' ) {
				// OCR0A(B)
				if( strncmp_P((char*)&tx_buf[1],PSTR("CR0"),3)==0 ) {
					if( tx_buf[4]=='A' )
						tos=0x36; //OCR0A;
					else if( tx_buf[4]=='B' )
						tos=0x3c; //OCR0B;
				}
			}
			else if( tx_buf[0]=='G' ) {
				// GIMSK
				if( strncmp_P((char*)&tx_buf[1],PSTR("IMSK"),4)==0 )
					tos=0x3b; //GIMSK;
			}
			tos+=0x20;
		}
		// register name 6 byte
		else if( namelen==6 ) {
			tos=0x15; //GPIOR2;
			if( tx_buf[0]=='P' ) {
				if( tx_buf[1]=='C' ) {
					//PCMSKn
					if( strncmp_P((char*)&tx_buf[2],PSTR("MSK"),3)==0 ) {
						if( tx_buf[5]=='0' ) tos=0x20; //PCMSK;
						else if( tx_buf[5]=='1' ) tos=0x04; //PCMSK1;
						else if( tx_buf[5]=='2' ) tos=0x05; //PCMSK2;
					}
				}
			}
			else if( tx_buf[0]=='T' ) {
				// TCCR0A(B)
				if( strncmp_P((char*)&tx_buf[1],PSTR("CCR0"),4)==0 ) {
					if( tx_buf[5]=='A' )
						tos=0x30; //TCCR0A;
					else if( tx_buf[5]=='B' )
						tos=0x33; //TCCR0B;
				}
			}
			tos+=0x20;
		} else
			tos = 0x15+0x20; //GPIOR2;
		// check bit flag
		if( uwptr!=namelen ) {
		  namelen++;
#ifdef __DEBUG__BITNAME__
		  for(uchar k=namelen;k<uwptr;k++) bbuff[k-namelen]=tx_buf[k];
#endif
		  if( (tx_buf[namelen]>='0')&&(tx_buf[namelen]<='7'))
		  	bport=(uchar)(tx_buf[namelen]-'0');
		} 
#ifdef __DEBUG__BITNAME__
			else bbuff[0]=0;
#endif
		namelen=0;
#endif
		}  // 	number
		wdt_reset();
        uwptr   = 0;

    } while(--recv_len);

	received = 0;
	recv_len=0;

    usbEnableAllRequests();

	if(iwptr>0) {
		rx_buf[iwptr]=0;
		usbSetInterrupt(rx_buf,8);
		if(iwptr>8) {
			while(!usbInterruptIsReady()) {
				wdt_reset();
				usbPoll();
			}
			usbSetInterrupt(rx_buf+8,8);
		}
		iwptr=0;
	}
}


static void hardwareInit(void)
{
uchar    i, j;

    /* activate pull-ups except on USB lines */
    USB_CFG_IOPORT   = (uchar)~((1<<USB_CFG_DMINUS_BIT)|(1<<USB_CFG_DPLUS_BIT));
    /* all pins input except USB (-> USB reset) */
#ifdef USB_CFG_PULLUP_IOPORT    /* use usbDeviceConnect()/usbDeviceDisconnect() if available */
    USBDDR    = 0;    /* we do RESET by deactivating pullup */
    usbDeviceDisconnect();
#else
    USBDDR    = (1<<USB_CFG_DMINUS_BIT)|(1<<USB_CFG_DPLUS_BIT);
#endif

    j = 0;
    while(--j){          /* USB Reset by device only required on Watchdog Reset */
        i = 0;
        while(--i)
        	wdt_reset();
    }

#ifdef USB_CFG_PULLUP_IOPORT
    usbDeviceConnect();
#else
    USBDDR    = 0;      /*  remove USB reset condition */
#endif
}


#ifdef __AVR_ATtiny4313__
ISR(PCINT0_vect)
{
	if( outbport!=8 ) {
		if( outdata & 1 )
			*outport |= (1<<outbport);
			else
			*outport &= ~(1<<outbport);
		outdata >>=1;
	}
}
ISR_ALIAS(PCINT1_vect,PCINT0_vect);
ISR_ALIAS(PCINT2_vect,PCINT0_vect);
#endif


int main(void)
{
#ifdef __AVR_ATtiny4313__
	cli();
	WDTCR |= 1<<WDP1;
	sei();
#endif
    odDebugInit();
    hardwareInit();
    usbInit();

    sei();
    for(;;){    /* main event loop */
        wdt_reset();
        usbPoll();

		if( usbInterruptIsReady() && (received != 0) )
		  Translate_buffer(inBuffer);
    }

    return 0;
}

